@@ -21,7 +21,7 @@ import json
 from builtins import str
 from collections import defaultdict
 from datetime import datetime, timedelta
-from typing import Any, DefaultDict, Dict, List, Optional, Set, Tuple, Union
+from typing import (Any, DefaultDict, Dict, List, Optional, Set, Tuple, Union, Iterator)
 
 import olm
 from jsonschema import SchemaError, ValidationError
@@ -31,7 +31,8 @@ from cachetools import LRUCache
 
 from . import (DeviceStore, GroupSessionStore, InboundGroupSession,
                InboundSession, OlmAccount, OlmDevice, OutboundGroupSession,
-               OutboundSession, Session, SessionStore, logger)
+               OutboundSession, Session, SessionStore, logger,
+               OutgoingKeyRequest)
 from ..api import Api
 from ..events import (BadEvent, BadEventType, Event,
                       ForwardedRoomKeyEvent, KeyVerificationAccept,
@@ -52,7 +53,6 @@ from ..store import MatrixStore
 from .key_export import decrypt_and_read, encrypt_and_save
 from .sas import Sas
 from ..event_builders import ToDeviceMessage, DummyMessage, RoomKeyRequestMessage
-from .sessions import OutgoingKeyRequest
 
 try:
     from json.decoder import JSONDecodeError
@@ -63,11 +63,17 @@ except ImportError:  # pragma: no cover
 DecryptedOlmT = Union[RoomKeyEvent, BadEvent, UnknownBadEvent, None]
 
 
+def chunks(lst, n):
+    """Yield successive n-sized chunks from lst."""
+    for i in range(0, len(lst), n):
+        yield lst[i:i + n]
+
+
 class KeyShareError(Exception):
     pass
 
 
-class Olm(object):
+class Olm:
     _olm_algorithm = 'm.olm.v1.curve25519-aes-sha2'
     _megolm_algorithm = 'm.megolm.v1.aes-sha2'
     _algorithms = [_olm_algorithm, _megolm_algorithm]
@@ -1484,6 +1490,16 @@ class Olm(object):
                         f"{event.sender} {event.device_id}.")
             self.key_re_requests_events[(device.user_id, device.device_id)].append(event)
 
+    def _decrypt_megolm_no_error(
+        self,
+        event: MegolmEvent,
+        room_id: str = None
+    ) -> Optional[Union[Event, BadEvent]]:
+        try:
+            return self.decrypt_megolm_event(event, room_id)
+        except EncryptionError:
+            return None
+
     def decrypt_megolm_event(self, event, room_id=None):
         # type (MegolmEvent, Optional[str]) -> Union[Event, BadEvent]
         room_id = room_id or event.room_id
@@ -1571,6 +1587,13 @@ class Olm(object):
             return bad
 
         parsed_dict["event_id"] = event.event_id
+
+        if "m.relates_to" not in parsed_dict["content"]:
+            try:
+                parsed_dict["content"]["m.relates_to"] = event.source["content"]["m.relates_to"]
+            except KeyError:
+                pass
+
         parsed_dict["sender"] = event.sender
         parsed_dict["origin_server_ts"] = event.server_timestamp
 
@@ -1724,6 +1747,19 @@ class Olm(object):
             room_id))
         self.create_outbound_group_session(room_id)
 
+    def should_share_group_session(self, room_id: str) -> bool:
+        """Should the client share a group session.
+
+        Returns True if no session was shared or the session expired, False
+        otherwise.
+        """
+        try:
+            session = self.outbound_group_sessions[room_id]
+        except KeyError:
+            return True
+
+        return session.expired or not session.shared
+
     def group_encrypt(
         self,
         room_id,  # type: str
@@ -1756,6 +1792,94 @@ class Olm(object):
 
         return payload_dict
 
+    def share_group_session_parallel(
+        self,
+        room_id: str,
+        users: List[str],
+        ignore_unverified_devices: bool = False
+    ) -> Iterator[Tuple[Set[Tuple[str, str]], Dict[str, Any]]]:
+        logger.info("Sharing group session for room {}".format(room_id))
+
+        if room_id not in self.outbound_group_sessions:
+            self.create_outbound_group_session(room_id)
+
+        group_session = self.outbound_group_sessions[room_id]
+
+        if group_session.shared:
+            self.create_outbound_group_session(room_id)
+            group_session = self.outbound_group_sessions[room_id]
+
+        key_content = {
+            "algorithm": self._megolm_algorithm,
+            "room_id": room_id,
+            "session_id": group_session.id,
+            "session_key": group_session.session_key,
+            "chain_index": group_session.message_index,
+        }
+
+        already_shared_set = group_session.users_shared_with
+        ignored_set = group_session.users_ignored
+
+        user_map = []
+        mark_as_ignored = []
+
+        for user_id in users:
+            for device in self.device_store.active_user_devices(user_id):
+                # No need to share the session with our own device
+                if device.id == self.device_id:
+                    ignored_set.add((user_id, device.id))
+                    continue
+
+                if self.is_device_blacklisted(device):
+                    ignored_set.add((user_id, device.id))
+                    continue
+
+                if ((user_id, device.id) in already_shared_set
+                        or (user_id, device.id) in ignored_set):
+                    continue
+
+                session = self.session_store.get(device.curve25519)
+
+                if not session:
+                    logger.warn(
+                        f"Missing Olm session for user {user_id} and device "
+                        f"{device.id}, skipping")
+                    continue
+
+                if not self.is_device_verified(device):
+                    if self.is_device_ignored(device):
+                        pass
+                    elif ignore_unverified_devices:
+                        mark_as_ignored.append(device)
+                    else:
+                        raise OlmUnverifiedDeviceError(
+                            device,
+                            f"Device {device.id} for user {device.user_id} is not "
+                            f"verified or blacklisted."
+                        )
+
+                user_map.append((user_id, device, session))
+
+        if mark_as_ignored:
+            self.store.ignore_devices(mark_as_ignored)
+
+        for user_map_chunk in chunks(user_map, self._maxToDeviceMessagesPerRequest):
+            to_device_dict = {"messages": {}}  # type: Dict[str, Any]
+            sharing_with = set()
+
+            for user_id, device, session in user_map_chunk:
+
+                olm_dict = self._olm_encrypt(session, device, "m.room_key",
+                                             key_content)
+                sharing_with.add((user_id, device.id))
+
+                if user_id not in to_device_dict["messages"]:
+                    to_device_dict["messages"][user_id] = {}
+
+                to_device_dict["messages"][user_id][device.id] = olm_dict
+
+            yield (sharing_with, to_device_dict)
+
     def share_group_session(
         self,
         room_id,  # type: str
